from Whittler.classes.NestedObjectPointer import NestedObjectPointerInterface
from Whittler.classes.MemoryCompressor import CompressedBytes
from tokenize import detect_encoding
from collections import defaultdict, OrderedDict
import hashlib
import re


class RelevanceInterface(NestedObjectPointerInterface):

    @property
    def relevant(self):
        return any(result.relevant for result in self.real_iter_values())

    def mark_irrelevant(self):
        for obj in self.real_iter_values():
            obj.mark_irrelevant()

    def mark_relevant(self):
        for obj in self.real_iter_values():
            obj.mark_relevant()

    def real_length(self):
        return len([e for e in self.real_iter_values()])

    def real_iter_values(self):
        raise NotImplementedError()


class Result(dict, RelevanceInterface):

    FRIENDLY_NAME = NotImplemented
    
    ATTRIBUTES = []
    SILENCED_ATTRIBUTES = set()
    SOLO_ATTRIBUTE = None
    SUPER_SOLO_ATTRIBUTE = None

    _init_run = False

    def __init__(self, resultdict=None):
        self.original_resultdict = resultdict
        self._cached_hash = 0
        self._cached_hash_valid = True
        self._frozen = False
        dict.__init__(self)
        self._relevant = True
        RelevanceInterface.__init__(self)

        if "whittler_filename" not in self.ATTRIBUTES:
            self.ATTRIBUTES.insert(0,"whittler_filename")
        if not resultdict is None:
            for k,v in resultdict.items():
                self[k] = v
        self._init_run = True

    def __setitem__(self, key, value):
        if type(value) not in (str, CompressedBytes):
            value = str(value)
        
        # This check is just for pickling/unpickling objects... the pickle implementation first calls __setitem__ with each of the
        # key/value pairs in this dict subclass, then sets the variable values above - this causes AttributeErrors to pop up because
        # self.ATTRIBUTES, self._frozen, etc have not been initialized yet.
        if not self._init_run:
            self.__init__()
        if key not in self.ATTRIBUTES:
            raise Exception(f"Unrecognized key {key} in this result.")
        if self._frozen:
            raise Exception("Cannot modify this result after it has been added to the database!")
        if self.Config.REMOVE_ANSI_CONTROL_CHARACTERS:
            value = self.filter_ansi(value)
        if self.Config.MEMORY_COMPRESSION:
            if self.Config.MemoryCompressor.training_mode:
                self.Config.MemoryCompressor.train(value, self)
            else:
                value = self.Config.MemoryCompressor.compress(value)
        self._cached_hash_valid = False
        super().__setitem__(key,value)
    
    def __getitem__(self, key):
        value = super().__getitem__(key)
        if type(value) == CompressedBytes:
            value = self.Config.MemoryCompressor.decompress(value)
        return value
    
    def __reduce__(self):
        cls = self.__class__
        args = ()
        state = self.__dict__.copy()
        list_iterator = None
        dict_iterator = iter({key:self.Config.MemoryCompressor.decompress(value) for key,value in dict(self).items()}.items())

        return (cls,args,state,list_iterator,dict_iterator)

    def __hash__(self):
        if self._cached_hash_valid:
            return self._cached_hash
        new_hashval = 0
        for attr in self.ATTRIBUTES:
            if attr in self:
                # TODO: faster to hash the compressed values if memory compression enabled, would that break anything?
                new_hashval = int(hashlib.md5(bytes(str(new_hashval)+attr+self[attr],'utf-8')).hexdigest(),16)
        self._cached_hash = new_hashval
        self._cached_hash_valid = True
        return new_hashval
    
    def __eq__(self, other):
        if not isinstance(other, Result):
            return False
        return hash(self) == hash(other)
    
    @staticmethod
    def give_result_dict_list(fname):
        raise NotImplementedError()
    
    # https://stackoverflow.com/questions/14693701/how-can-i-remove-the-ansi-escape-sequences-from-a-string-in-python
    ansi_escape = re.compile(r'\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])')
    
    # Takes a unicode string, and filters out all 7-bit ANSI escape characters.
    def filter_ansi(self, s):
        return self.ansi_escape.sub('', s)
    
    # "Just read the f*cking string from the file already!"
    @staticmethod
    def force_read_file_to_string(fname):
        try:
            with open(fname, "r",encoding=detect_encoding(fname)) as f:
                return f.read()
        except (UnicodeDecodeError,UnicodeError):
            with open(fname, "rb") as f:
                b = f.read()
            b.replace(b"\x00",b"\\x00")
            b = b"".join([(bytes(bt) if int(bt) < 0x80 else bytes("\\x{:0>2}".format(hex(int(bt))[2:]),'ascii')) for bt in b])
            return b.decode("ascii")
    
    # This returns an elegant representation of the result. By default, it does not print the "diff"
    # component itself because it is usually very long and clutters up the output.
    def pretty_repr(self):
        s = ""
        def adds(content):
            nonlocal s
            s += str(content) + "\n"
        for k,v in self.items():
            if not self.SUPER_SOLO_ATTRIBUTE is None:
                if k != self.SUPER_SOLO_ATTRIBUTE:
                    continue
                adds(v)
                break
            elif not self.SOLO_ATTRIBUTE is None:
                if k != self.SOLO_ATTRIBUTE:
                    continue
            if not v.strip():
                continue
            adds(k)
            if k in self.SILENCED_ATTRIBUTES and k != self.SOLO_ATTRIBUTE:
                adds("... silenced ...")
                adds("")
                continue
            adds("\n".join("    "+line for line in v.splitlines()))
            adds("")
        return s
    
    def display(self):
        print(self.pretty_repr())
    

    #######################################
    #  RelevanceInterface implementations
    #

    @property
    def relevant(self):
        return self._relevant
    
    def mark_irrelevant(self):
        self._relevant = False

    def mark_relevant(self):
        self._relevant = True
    
    def real_iter_values(self):
        return [self]
    

    #################################################
    #  NestedObjectPointerInterface implementations
    #
    
    def enumerate_child_pointers(self, pointer_to_me):
        return OrderedDict({str(hash(self)):pointer_to_me})
    
    def size(self):
        return 1
    
    def all_result_objects(self):
        return [self]

    def show_view(self, pointer_to_me=None, ct=0, limit=None, show_irrelevant=False, sort_by=None, sort_numeric=False, sort_reverse=False):
        return (self.pretty_repr(), pointer_to_me)
    
    def export(self):
        data = {attr:self[attr] for attr in self.ATTRIBUTES}
        return {**data, "whittler_filename":self["whittler_filename"]}
    

class RelevanceFilteredResultList(list, RelevanceInterface):
    def __init__(self,iterable=None):
        list.__init__(self)
        RelevanceInterface.__init__(self)
        if not iterable is None:
            for e in iterable:
                self.append(e)
    
    def __setitem__(self, index, value):
        assert isinstance(value, RelevanceInterface)
        list.__setitem__(self,index,value)
    
    def __iter__(self):
        yield from (e for e in self.real_iter_values() if e.relevant)
    
    def __len__(self):
        return len([e for e in self])
    
    # todo: speed up "result in RFRLobj" lookups - most of the time is spent checking "if result in self.results" in ResultDatabase.add_result
    
    def append(self, newitem):
        assert isinstance(newitem, RelevanceInterface)
        list.append(self,newitem)
    
    def extend(self, iterable):
        for e in iterable:
            self.append(e)
    
    def yield_irrelevant(self):
        yield from (e for e in list.__iter__(self) if not e.relevant)
    
    def get_by_id(self, result_id):
        return list(filter(self.real_iter_values(),key=lambda result:str(hash(result))[:6] == result_id))[0]
    

    #######################################
    #  RelevanceInterface implementations
    #
    
    def real_iter_values(self):
        yield from list.__iter__(self)


    #################################################
    #  NestedObjectPointerInterface implementations
    #
    
    def enumerate_child_pointers(self, pointer_to_me):
        ret = OrderedDict()
        for i in range(self.real_length()):
            if self[i].relevant:
                if not pointer_to_me is None:
                    ptr = pointer_to_me.copy()
                    ptr.get_by_index(i)
                    ret[i] = ptr
                else:
                    ret[i] = None
        if not pointer_to_me is None:
            self._cached_pointers = ret
        return ret
    
    def size(self):
        return len(self)
    
    def all_result_objects(self):
        return [result for result in self]

    _numeric_tokenize = re.compile(r'(\d+)|(\D+)').findall
    def _numeric_sortkey(self,str):
        return tuple(int(num) if num else alpha for num,alpha in self._numeric_tokenize(str))

    def show_view(self, pointer_to_me=None, ct=0, limit=None, show_irrelevant=False, sort_by=None, sort_numeric=False, sort_reverse=False):
        ret = OrderedDict()
        s = ""
        def adds(string, prefix=""):
            nonlocal s
            s += prefix + string + "\n"
        child_ptrdict = self.give_child_pointers(pointer_to_me)
        if not sort_by is None:
            # "reversed" actually yields the expected (i.e. non-reversed-seeming) result...
            reverse = lambda iter: lambda iter: reversed(iter) if sort_reverse else iter
            if sort_numeric:
                child_ptrdict = OrderedDict(reverse(sorted(child_ptrdict.items(), key=lambda tup: self._numeric_sortkey(self[tup[0]][sort_by]))))
            else:
                child_ptrdict = OrderedDict(reverse(sorted(child_ptrdict.items(), key=lambda tup: self[tup[0]][sort_by])))
        for key, ptr in child_ptrdict.items():
            result = self[key]
            if not show_irrelevant and not result.relevant:
                continue
            result_id = str(hash(result))[:6]
            if not result.SUPER_SOLO_ATTRIBUTE is None:
                adds(result.pretty_repr())
            else:
                adds("~~~~~~~~~~~~~~~~~~~~~~~~~")
                adds(f"  Result with ID {result_id}    ")
                adds("~~~~~~~~~~~~~~~~~~~~~~~~~")
                for line in result.pretty_repr().split("\n"):
                    adds(line, prefix="| ")
                s += "\n\n"
            ret[int(result_id)] = ptr
            if not limit is None and len(ret) >= limit:
                break
        return (s, ret)
    
    def export(self):
        return [result.export() for result in self]
    

class ValueLengthSortedResultDict(defaultdict, RelevanceInterface):

    def __init__(self, *args, **kwargs):
        defaultdict.__init__(self, *args, **kwargs)
        RelevanceInterface.__init__(self)
        NestedObjectPointerInterface.__init__(self)
    
    def items(self):
        yield from reversed(sorted(defaultdict.items(self),key=lambda tup:len(tup[1])))
        
    def keys(self):
        yield from (k for k,v in self.items())
        
    def values(self):
        yield from (v for k,v in self.items())
    
    def __setitem__(self, key, value):
        assert isinstance(value, RelevanceInterface)
        defaultdict.__setitem__(self,key,value)
    
    def __len__(self):
        return len([obj for obj in self.values() if obj.relevant])
    

    #######################################
    #  RelevanceInterface implementations
    #
    
    def real_iter_values(self):
        yield from (self[key] for key in defaultdict.__iter__(self))
    

    #################################################
    #  NestedObjectPointerInterface implementations
    #
    
    def enumerate_child_pointers(self, pointer_to_me):
        ret = OrderedDict()
        for k in self.keys():
            ptr = pointer_to_me.copy()
            ptr.get_by_index(k)
            ret[k] = ptr
        self._cached_pointers = ret
        return ret
    
    def size(self):
        return sum(obj.size() for obj in self.values())
    
    def all_result_objects(self):
        ret = []
        for resultcontainer in self.values():
            ret.extend(resultcontainer.all_result_objects())
        return ret
    
    _numeric_tokenize = re.compile(r'(\d+)|(\D+)').findall
    def _numeric_sortkey(self,str):
        return tuple(int(num) if num else alpha for num,alpha in self._numeric_tokenize(str))

    def show_view(self, pointer_to_me=None, ct=0, limit=None, show_irrelevant=False, sort_by=None, sort_numeric=False, sort_reverse=False):
        max_width = self.Config.MAX_OUTPUT_WIDTH
        ret = OrderedDict()
        
        child_pointers = self.give_child_pointers(pointer_to_me)
        child_repr_values = OrderedDict()
        for key, ptr in child_pointers.items():
            value = self[key]
            sanitized_key = key.strip().replace("\n","\\n").replace("\r","\\r") # todo: sanitize more thoroughly
            child_repr_values[key] = (ct,value.real_length(),len(value),sanitized_key)
        # "reversed" actually yields the expected (i.e. non-reversed-seeming) result...
        reverse = lambda iter: reversed(iter) if sort_reverse else iter
        if sort_by == "total":
            ordered_keys = dict(reverse(sorted(child_repr_values.items(), key=lambda tup: tup[1][1]))).keys()
        elif sort_by == "relevant":
            ordered_keys = dict(reverse(sorted(child_repr_values.items(), key=lambda tup: tup[1][2]))).keys()
        elif sort_by == "attribute value":
            if sort_numeric:
                ordered_keys = dict(reverse(sorted(child_repr_values.items(), key=lambda tup: self._numeric_sortkey(tup[1][3])))).keys()
            else:
                ordered_keys = dict(reverse(sorted(child_repr_values.items(), key=lambda tup: tup[1][3]))).keys()
        else:
            ordered_keys = child_repr_values.keys()
        
        lines = []
        for key in ordered_keys:
            value = self[key]
            ptr = child_pointers[key]
        # for key, ptr in self.give_child_pointers(pointer_to_me).items():
        #     value = self[key]
            if not show_irrelevant and not len(value):
                continue
            sanitized_key = key.strip().replace("\n","\\n").replace("\r","\\r") # todo: sanitize more thoroughly
            line = "| {:4d} | {:5d} | {:8d} | {}".format(ct,value.real_length(),len(value),sanitized_key)
            if len(line) > max_width-2:
                line = line[:max_width-2-7]+" [...] "
            lines.append(line)
            ret[ct] = ptr
            ct += 1
            if not limit is None and len(lines) >= limit:
                break
        headertext = "|  id  | total | relevant | attribute value  "
        max_linelen = max(len(line) for line in lines) if len(lines) else 0
        max_linelen = len(headertext) if len(headertext) > max_linelen else max_linelen
        header_footer = "+" + "="*(max_linelen-1) + "+\n"
        divider = "|{:-<6}+{:-<7}+{:-<10}+".format("","","")
        divider = divider + "-"*(max_linelen-len(divider)) + "|\n"
        header = header_footer
        header += headertext + " "*(max_linelen-len(headertext)) + "|\n"
        header += divider
        s = header
        for line in lines:
            s += line + " "*(max_linelen-len(line)) + "|\n"
        s += header_footer
        return (s, ret)
    
    def export(self):
        ret = []
        for value in self.values():
            if value.relevant:
                ret.extend(value.export())
        return ret
    

class ResultDict(ValueLengthSortedResultDict):
    def __init__(self, parent_rdc=None):
        super().__init__(RelevanceFilteredResultList)
        self.parent_rdc = parent_rdc


class ResultDictContainer(ValueLengthSortedResultDict):
    def __init__(self, resultdb):
        super().__init__(lambda: ResultDict(self))
        self.resultdb = resultdb



from Whittler.classes.Result import Result
from Whittler.config import Config
import json

class BanditResult(Result):

    Config = Config.copy({"FILE_ENCODING":"utf-8"})

    FRIENDLY_NAME = "bandit"
    
    ATTRIBUTES = [
      "code",
      "filename",
      "issue_confidence",
      "issue_severity",
      "issue_text",
      "line_number",
      "line_range",
      "more_info",
      "test_id",
      "test_name",
    ]

    @staticmethod
    def give_result_dict_list(fname):
        with open(fname,"r", encoding=BanditResult.Config.FILE_ENCODING) as f:
            output = json.loads(f.read())
            return output["results"]


from Whittler.classes.Result import Result
import re
import json


class TrufflehogResult(Result):

    FRIENDLY_NAME = "trufflehog"
    
    ATTRIBUTES = [
        "branch",
        "commit",
        "commitHash",
        "date",
        "diff",
        "path",
        "printDiff",
        "reason",
        "stringsFound"
    ]

    @staticmethod
    def give_result_dict_list(fname):
        with open(fname,"r", encoding=TrufflehogResult.Config.FILE_ENCODING) as f:
            output_raw = f.read()
            # For some reason, trufflehog just tacks together JSON objects without commas or an array to contain them,
            # e.g. {blah:blah} {blah:blah} . We want to turn that into [{blah:blah},{blah:blah}] .
            output_raw = re.sub(r"}\s*{","},{",output_raw)
            output_raw = f"[{output_raw}]"
            return json.loads(output_raw)


from Whittler.classes.Result import Result
from Whittler.config import Config
import json
from collections import defaultdict
# try:
#     import jsonschema
# except ImportError:
#     print("The jsonschema module is required to perform sarif parsing. Try running 'python -m pip install jsonschema'")
#     raise


class SarifResult(Result):

    Config = Config.copy({"FILE_ENCODING":"utf-8"})

    FRIENDLY_NAME = "sarif"
    
    ATTRIBUTES = []

    @staticmethod
    def collapse_recursive_json(js:dict,prefix=[],ct=""):
        ret = {}
        if not isinstance(js, dict):
            ret[".".join(prefix)] = str(js)
            return ret
        for k,v in js.items():
            if isinstance(v,list):
                prefix.append(k)
                ct = 1
                for e in v:
                    ret = {**ret,**SarifResult.collapse_recursive_json(e,ct=ct)}
                    ct += 1
                prefix.pop()
            elif isinstance(v,dict):
                prefix.append(f"{k}{ct}")
                new_dict = SarifResult.collapse_recursive_json(v)
                ret = {**ret,**new_dict}
                prefix.pop()
            else:
                ret[".".join([*prefix,k])] = v
        return ret

    @staticmethod
    def give_result_dict_list(fname):
        if not fname.endswith(".sarif"):
            print(f"Warning: ignoring {fname} as its filename does not end with \".sarif\".")
            return []
        with open(fname,"r", encoding=SarifResult.Config.FILE_ENCODING) as f:
            sarif_json = json.loads(f.read())
        
        # I'd prefer to have this be a set(), but it needs to stay ordered for convenience.
        all_keys = []

        run = sarif_json["runs"][0]
        
        version = sarif_json["version"]

        # Get the rules, collapsed recursively, as a list of rule dictionaries (version-dependent)
        if version == "2.1.0":
            rules = [SarifResult.collapse_recursive_json(rule) for rule in run["tool"]["driver"]["rules"]]
            
        elif version == "1.0.0":
            rules = [SarifResult.collapse_recursive_json(rule) for rule in run["rules"].values()]
        else:
            print(f"Warning: Unsupported SARIF version v{version}")
            return []
        
        # Sort rules into a dict indexed by rule ID
        rules = {rule["id"]:rule for rule in rules}
        for rule in rules.values():
            del rule["id"]
        
        # Add the union of all rule attributes into the all_keys set (prepending "rules.")
        for r in rules.values():
            for k in r.keys():
                key = f"rules.{k}"
                if key not in all_keys:
                    all_keys.append(key)
        
        # get the results, collapsed recursively, as a list of result dictionaries
        ret = [SarifResult.collapse_recursive_json(result) for result in run["results"]]
        if version == "2.1.0" or version == "1.0.0":
            results = []
            for rawresult in (SarifResult.collapse_recursive_json(result) for result in ret):
                result = {}
                regiondict = defaultdict(dict)
                for k,v in rawresult.items():
                    if "region" in k:
                        klist = k.split(".")
                        region_key_list = []
                        i = 0
                        while "region" not in klist[i]:
                            region_key_list.append(klist[i])
                            i += 1
                        region_key_list.append(klist[i])
                        region_key = ".".join(region_key_list)
                        regiondict[region_key][".".join(klist[i+1:])] = v
                    else:
                        result[k] = v
                for k,v in regiondict.items():
                    result[k] = str(dict(sorted(v.items())))
                results.append(result)
            ret = results

        # add the union of all result attributes into the all_keys set
        for r in ret:
            for k in r.keys():
                if k not in all_keys:
                    all_keys.append(k)
        
        # set this class' ATTRIBUTES to reflect all the keys we've collected - otherwise we'll have errors
        # when casting the returned dictionaries to SarifResult objects
        for k in all_keys:
            if k not in SarifResult.ATTRIBUTES:
                SarifResult.ATTRIBUTES.append(k)

        for r in ret:
            # get the rule corresponding to this result
            rule = rules[r["ruleId"]]
            
            # add each rule attribute to this result, prepending "rules."
            for k,v in rule.items():
                r[f"rules.{k}"] = v
            
            # If a key in ATTRIBUTES is not present, then add it as an empty string - every result dictionary
            # returned must have every attribute specified in this class' ATTRIBUTES variable
            for k in SarifResult.ATTRIBUTES:
                if k not in r:
                    r[k] = ""
        return ret





from collections import namedtuple, OrderedDict
from Whittler.config import ConfigurableInterface

Vertex = namedtuple("Vertex", ["operation","value"])

class NestedObjectPointer:
    def __init__(self, base_object):
        self.base_object = base_object
        # A list of Vertex objects that will be used to find the pointed object from the base object.
        self.path = []
        self._pointed_object = None
    
    def is_base_pointer(self):
        return not len(self.path)

    def access_property(self, propname):
        self.path.append(Vertex("__getattribute__", propname))
        self.flush_object_cache()
    
    def get_by_index(self, key):
        self.path.append(Vertex("__getitem__", key))
        self.flush_object_cache()
    
    def give_pointed_object(self):
        if not self._pointed_object is None:
            return self._pointed_object
        o = self.base_object
        for operation, value in self.path:
            o = getattr(o,operation)(value)
        self._pointed_object = o
        return o
    
    def flush_object_cache(self):
        self._pointed_object = None
    
    def go_up_level(self):
        self.flush_object_cache()
        return self.path.pop()
    
    def __eq__(self,other):
        if not isinstance(other, self.__class__):
            return False
        if self.base_object is not other.base_object:
            return False
        if len(self.path) != len(other.path):
            return False
        for i in range(len(self.path)):
            if self.path[i] != other.path[i]:
                return False
        return True
    
    def __repr__(self):
        return " -> ".join(["base"]+[str(vertex.value) for vertex in self.path])
    
    def copy(self):
        ret = type(self)(self.base_object)
        ret.path = [e for e in self.path]
        ret._pointed_object = self._pointed_object
        return ret


class NestedObjectPointerInterface(ConfigurableInterface):
    def __init__(self):
        self._cached_pointers = None
    
    def flush_pointer_cache(self):
        self._cached_pointers = None
    
    def give_child_pointers(self, pointer_to_me):
        if not self._cached_pointers is None:
            return self._cached_pointers
        return self.enumerate_child_pointers(pointer_to_me)
    
    def enumerate_child_pointers(self, pointer_to_me):
        raise NotImplementedError()

    def size(self):
        raise NotImplementedError()

    def all_result_objects(self):
        raise NotImplementedError()
    
    def show_view(self, pointer_to_me=None, ct=0, limit=None, show_irrelevant=False, sort_by=None, sort_numeric=False, sort_reverse=False):
        raise NotImplementedError()

    def export(self):
        raise NotImplementedError()


# ParentRelationship = namedtuple("ParentRelationship",["parent","vertex"])

# class NestedObjectPointerInterface:
#     def __init__(self):
#         self._parent_pointers = {}

#     def declare_parent_relationship(self, parent, vertex):
#         parent_relationship = ParentRelationship(parent,vertex)
#         self._parent_pointers[parent_relationship] = self.give_pointers_to_self(parent_relationship)

#     def give_pointers_to_self(self, parent_relationship=None):
#         if parent_relationship is None:
#             return list(self._parent_pointers.values())
#         if parent_relationship in self._parent_pointers:
#             return self._parent_pointers[parent_relationship]
        
#         parents = [parent_relationship["parent"]]
#         if not parents:
#             return NestedObjectPointer(self)
#         pointers = parent.give_pointers_to_self()
#         pointer.path.append(self.give_parent_to_self_vertex())
#         return pointer


# class ResultDatabasePointer:
#     all_result_sets = [
#         "categorized",
#         "grouped"
#     ]
#     def __init__(self, resultdb):
#         self.resultdb = resultdb
#         self.resultset = None
#         self.pointer = []
    
#     def switch_to_resultset(self, resultset=None):
#         assert resultset is None or resultset in self.all_result_sets
#         self.resultset = resultset
    
#     def switch_to_root_context(self):
#         self.switch_to_resultset()
#         self.pointer = []
    
#     def give_pointed_object(self):
#         if self.resultset is None:
#             return self.resultdb
#         context = getattr(self.resultdb, self.resultset)
#         for ptr in self.pointer:
#             context = context[ptr]
#         return context
    
#     def switch_to_parent_context(self):
#         assert self.resultset is not None
#         if not len(self.pointer):
#             self.resultset = None
#             return
#         self.pointer.pop()
    
#     def switch_to_deeper_context(self,key_or_resultset):
#         if self.resultset is None:
#             self.switch_to_resultset(key_or_resultset)
#         else:
#             self.pointer.append(key_or_resultset)
    
#     def copy(self):
#         ret = type(self)(self.resultdb)
#         ret.resultset = self.resultset
#         ret.pointer = [e for e in self.pointer]
#         return ret



from Whittler.classes.MemoryCompressor import MemoryCompressor

class Config:
    # The encoding of the files that will be parsed by Whittler.
    FILE_ENCODING = 'utf-8'

    # Some tools put ANSI control characters to color their output, which is nice in the console, but pesky
    # when we're trying to do bulk operations with them.
    REMOVE_ANSI_CONTROL_CHARACTERS = False

    # Grouping is performed by combining the similarity ratings from the Damerau-Levenshtein algorithm and the
    # Ratcliff-Obershelp algorithm using a Sum-Of-Squares method, except instead of precisely squaring/sqrting,
    # the exponent is modifiable through the SIMILARITY_EXPONENT parameter below.
    SIMILARITY_THRESHOLD = 0.5

    # When the SIMILARITY_THRESHOLD is auto-modified, it is modified using the following formula:
    # SIMILARITY_THRESHOLD += (DESIRED - SIMILARITY_THRESHOLD) / SIMILARITY_THRESHOLD_MODIFICATION_FACTOR
    # where DESIRED is either 1 or 0 based on whether we want to increase or decrease the threshold.
    SIMILARITY_THRESHOLD_MODIFICATION_FACTOR = 5

    # The results from the Damerau-Levenshtein algorithm and the Ratcliff-Obershelp algorithm are combined using
    # a sum-of-squares with this exponent. That is, similarity=(dl_similarity**exp + ro_similarity**exp)**(1/exp),
    # where both dl_similarity and ro_similarity are scaled from 0 to 1, with 1 being the most similar.
    SIMILARITY_EXPONENT = 2

    # The maximum printing width of the console, when printing tables and a few other things.
    MAX_OUTPUT_WIDTH = 100

    # Decreasing verbosity will suppress some of the output.
    VERBOSITY = 3

    # Enabling transparent in-memory compression can enable working with much larger datasets in memory, at the
    # cost of some performance.
    MEMORY_COMPRESSION = False


    ##################

    MEMORY_COMPRESSION_TRAIN_COUNT = 1000

    MemoryCompressor = MemoryCompressor()

    def __getstate__(self):
        state = self.__dict__.copy()
        if "MemoryCompressor" in state:
            del state["MemoryCompressor"]
        return state
    
    def __setstate__(self, state):
        self.__dict__.update(state)

    @classmethod
    def copy(cls, overridden_values={}):
        return type(f"_{cls.__name__}",(cls,),overridden_values)

class ConfigurableInterface:
    Config = Config


from Whittler.classes.Result import Result
# from Whittler.config import Config

class SampleResult(Result):

    # # This is totally optional, but you can override configuration settings as follows. See config.py for
    # # more information on the configuration settings.
    # Config = Config.copy({"FILE_ENCODING":"utf-8"})

    # This specifies the name that the user can specify when initializing Whittler, e.g. python Whittler.py --config samplemod
    FRIENDLY_NAME = "samplemod"
    
    # This is a list of all attributes that this Result object is expected to (and is able to) contain. Result objects
    # are basically dicts with a bunch of additional functionality, so this is like a list of allowed values for the
    # Result dict's keys.
    ATTRIBUTES = [
        "full name",
        "address",
        "phone number"
    ]

    # This method (which must be declared as a staticmethod as below) is expected to take a filename, and return a list
    # of dicts, each of which will be casted to an object of this class' type (in this case, SampleResult). These dicts
    # MUST have a string as both a key and as a value - I hope to allow recursively nested structures in the future, but
    # for now only string key/value pairs are supported. An example of a valid return value would be:
    # [
    #     {
    #         "full name" : "Chuntley Bortler",
    #         "address" : "1 Failure Ave, Falterton WY 01324",
    #         "phone number": "666-420-6969"
    #     },
    #     {
    #         "full name" : "Flubbles Troughslogger",
    #         "address" : "42 Nihilism St, Derridaville KS 80085",
    #         "phone number": "555-867-5309"
    #     }
    # ]
    @staticmethod
    def give_result_dict_list(fname):
        with open(fname, "r") as f:
            # Do whatever you need to do here to convert the content of the file in fname to the expected format.
            # This import should be at the top, but I didn't want to complicate the imports in the example module.
            import json
            output = json.loads(f.read())
            return output["people"]


from Whittler.classes.Result import Result


class PlainTextResult(Result):

    FRIENDLY_NAME = "plaintext"
    
    ATTRIBUTES = [
       "line",
       "line length"
    ]

    @staticmethod
    def give_result_dict_list(fname):
        with open(fname,"r", encoding=PlainTextResult.Config.FILE_ENCODING) as f:
            return [{"line":line,"line length":len(line)} for line in f.readlines()]


from Whittler.classes.NestedObjectPointer import NestedObjectPointer
from Whittler.config import Config
import textwrap
import random


actions = {
    "navigation" : {
        "show [[limit]]" : "Show the current data context, up to [limit] entries (shows all entries by default). Mutes "+\
                           "results or table entries with 0 relevant results.",
        "showall [[limit]]" : "Show the current data context, up to [limit] entries (shows all entries by default). Includes "+\
                              "results or table entries with 0 relevant results.",
        "dig [attr]" : "Dig into a specific data grouping category, either by attribute name, or by attribute id",
        "up" : "Dig up a level into the broader data grouping category",
        "top" : "Dig up to the top level",
        "dump [[limit]]" : "Display every relevant result in every category, up to [limit] entries (shows all by default)",
        "dumpall [[limit]]" : "Display every result, both relevant and irrelevant, in every category, up to [limit] entries "+\
                              "(shows all by default)",
        #"search [searchstr]" : "Finds the thing", # todo
        "exit" : "Gracefully exit the program"
    },
    "data model interaction" : {
        "irrelevant [[id]]" : "Mark all elements in the current context, or those referenced by [id], as irrelevant results",
        "relevant [[id]]" : "Mark all elements in the current context, or those referenced by [id], as relevant results",
        "group [id] [[attr]]" : "Using data science, group all results in the database by similarity to the attribute referenced "+\
                                "by [id]. Or, if [id] points to a specific result, group by similarity to a specific attribute of "+\
                                "the result referenced by [id].",
        "game [[id]]" : "Play a game where individual results are presented one-by-one, and the user is asked whether "+\
                        "the result is relevant or not and why. Using this information, other similar results are also "+\
                        "eliminated in bulk. If [id] is specified, then the results presented are limited to the result "+\
                        "group represented by the specified [id].",
        "filter [str] [[attr]]" : "Mark all results containing [str] in a particular attribute as irrelevant (case-insensitive)"
    },
    "output" : {
        "quiet [[attr]]" : "Suppress an attribute from being displayed when printing out raw result data",
        "unquiet [[attr]]" : "Undo the suppression from an earlier quiet command",
        "solo [[attr]]" : "Print only a single attribute's value when printing out raw result data",
        "SOLO [[attr]]" : "Print ONLY a single attribute's value when printing out raw result data, with no context IDs or "+\
                          "attribute value names",
        "unsolo" : "Disable solo mode. Note that this retains any attributes suppressed using the \"quiet\" command.",
        "sort [colname]" : "Sorts the displayed results by the value in the specified column. Use quotes if the column name "+\
                           "has a space in it.",
        "sortn [colname]" : "Sorts the displayed results numerically by the value in the specified column. Use quotes if the "+\
                            "column name has a space in it.",
        "rsort [colname]" : "Reverse-sorts the displayed results by the value in the specified column. Use quotes if the column name "+\
                            "has a space in it.",
        "rsortn [colname]" : "Reverse-sorts the displayed results numerically by the value in the specified column. Use quotes if the "+\
                             "column name has a space in it.",
        "history" : "Print all commands that have been run in this session so far",
        "width [numchars]" : "Modify the maxiumum terminal width, in characters, that all output will be formatted to",
        "exportjson [fname] [[id]]" : "Export all relevant results in JSON form at into the file [fname]. Optionally, limit "+\
                                      "the output to the result set as referenced by [id].",
        "export [fname] [[id]]" : "Export all relevant results in Python Pickle (serialized binary) form at into the file [fname]. "+\
                                  "Optionally, limit the output to the result set as referenced by [id]."
    }
}

# A simple history of the commands that have been run in the Whittler shell.
command_history = []

# The file to which the command history will be written, if any. By default, a .txt file named after
# the current date and time, placed in the .whittler directory in the user's home folder.
command_redirect_file = None
def set_command_redirect_file(obj):
    global command_redirect_file
    command_redirect_file = obj
def get_command_redirect_file():
    return command_redirect_file

# The file to which all Whittler output will be written, if any. By default, a .txt file named after
# the current date and time, placed in the .whittler directory in the user's home folder.
global_redirect_file = None
def set_global_redirect_file(obj):
    global global_redirect_file
    global_redirect_file = obj
def get_global_redirect_file():
    return global_redirect_file

# The file to which the output of the specified command will be written. This adds support for shell-
# style redirection, e.g. "showall > ./my_file.txt"
redirect_file = None
def set_redirect_file(obj):
    global redirect_file
    redirect_file = obj
def get_redirect_file():
    return redirect_file


class _NoInput:
    pass
def wprint(s=_NoInput,end="\n"):
    if s is _NoInput:
        s = ""
    if not redirect_file is None:
        redirect_file.write(str(s)+end)
    if not global_redirect_file is None:
        global_redirect_file.write(str(s)+end)
    print(s,end=end)

def winput(msg):
    user_input = input(msg)
    command_history.append(user_input)
    if not command_redirect_file is None:
        command_redirect_file.write(user_input+"\n")
    if not redirect_file is None:
        redirect_file.write(str(msg))
        redirect_file.write(user_input+"\n")
    if not global_redirect_file is None:
        global_redirect_file.write(str(msg))
        global_redirect_file.write(user_input+"\n")
    return user_input

def print_help():
    wprint()
    printstr = "|   {{:<{}}}  :  "
    for actioncategory, actiondict in actions.items():
        longest_action_name = max(len(action) for action in actiondict.keys())
        wprint(f"{actioncategory}:")
        for actionname, action_description in actiondict.items():
            prologue = printstr.format(longest_action_name).format(actionname)
            description = textwrap.wrap(action_description, Config.MAX_OUTPUT_WIDTH-len(prologue))
            for line in description:
                wprint(f"{prologue}{line}")
                prologue = "|" + " "*(len(prologue)-1)
        wprint()
    wprint("NOTE: This shell supports quoted arguments and redirecting command output to a file using the \">\" operator.")
    wprint()

# False means no value was supplied for this arg position, None means failed to parse as int
def get_int_from_args(args, position=0):
    if not len(args) > position:
        return False
    else:
        try:
            argval = int(args[position])
        except ValueError:
            return None
    return argval

# False means no value was supplied for this arg position, None means failed to find corresponding context pointer
def get_ptr_from_id_arg(resultdb, args, id_arg_position=0, quiet=False):
    if not isinstance(resultdb.context_pointers, dict):
        if not quiet:
            wprint(f"Can't dig deeper.")
        return None
    
    choice = get_int_from_args(args, position=id_arg_position)
    
    # No value was supplied for this arg position
    if choice is False:
        return False
    
    # we couldn't parse it as an int, so it must be a string literal attribute value
    if choice is None:
        choice = args[id_arg_position]
        for ptr in resultdb.context_pointers.values():
            # The attribute values will be the .value property of the Vertex object given by the last
            # get_by_index operation called on this pointer.
            if ptr.path[-1].value == choice:
                return ptr
        if not quiet:
            wprint(f"Could not recognize \"{choice}\" as one of the attributes of this dataset.\n")
        return None
    
    # The value supplied was an int, so should be looked up in the context_pointers dict
    elif choice not in resultdb.context_pointers:
        if not quiet:
            wprint(f"Could not recognize {choice} as one of the IDs listed above.\n")
        return None
    return resultdb.context_pointers[choice]

# False means no value was supplied for this arg position, None means failed to find corresponding context pointer
def get_attrname_from_attribute_arg(resultdb, args, attr_arg_position=0):
    ptr = get_ptr_from_id_arg(resultdb, args, id_arg_position=attr_arg_position, quiet=True)
    if ptr is False:
        return False
    attrname = args[attr_arg_position]
    if attrname not in resultdb.result_class.ATTRIBUTES:
        return None
    return attrname

def select_attribute(resultdb, msg):
    attrs = resultdb.result_class.ATTRIBUTES
    for i in range(len(attrs)):
        wprint(f" {i} : {attrs[i]}")
    wprint()
    try:
        index = int(winput(msg))
    except ValueError:
        return None
    return attrs[index]

def group_interactive(resultdb, groupattr, groupval):
    wprint("Creating a result group set based on this attribute value...")
    SIMILARITY_THRESHOLD = resultdb.Config.SIMILARITY_THRESHOLD
    while True:
        print(SIMILARITY_THRESHOLD)
        similar_results = resultdb.find_similar_results(groupattr, groupval, similarity_threshold=SIMILARITY_THRESHOLD)
        wprint()
        for val in sorted(set([res[groupattr] for res in similar_results])):
            wprint(f"   {val.strip()}")
        wprint()
        wprint(f"I found {len(similar_results)} similar results with \"{groupattr}\" values as shown above.")
        wprint("I want to create a result group set based on these findings, but I want to make sure it's OK.")
        while True:
            grouping_choice = winput("Is this grouping OK (1), too conservative (2), or too liberal (3)? Or, just abandon the group creation (4)? ")
            bad_input = False
            try:
                grouping_choice = int(grouping_choice)
                if grouping_choice not in [1,2,3,4]:
                    bad_input = True
            except:
                bad_input = True
            if bad_input:
                wprint("I didn't understand your input.")
                continue
            break
        if grouping_choice == 1:
            pass
        elif grouping_choice == 2:
            SIMILARITY_THRESHOLD -= SIMILARITY_THRESHOLD / Config.SIMILARITY_THRESHOLD_MODIFICATION_FACTOR
            continue
        elif grouping_choice == 3:
            SIMILARITY_THRESHOLD += (1 - SIMILARITY_THRESHOLD) / Config.SIMILARITY_THRESHOLD_MODIFICATION_FACTOR
            continue
        elif grouping_choice == 4:
            wprint("OK, abandoning group creation.")
            break
        resultdb.register_grouped_results(groupattr, groupval, similar_results)
        wprint("Created result group based on this entry. You can view this group in the \"groups\" pane in the top context.")
        ptr = NestedObjectPointer(resultdb)
        ptr.access_property("grouped_results")
        ptr.get_by_index(groupval)
        return ptr

def parse_user_input(user_input):
    if not user_input.strip():
        return None
    
    # Support strings with escaped quote characters
    user_input = user_input.replace("\\'","###_TRUFFLEHOG_SINGLE_QUOTE_###")
    user_input = user_input.replace("\\\"","###_TRUFFLEHOG_DOUBLE_QUOTE_###")

    # Support multiple commands in a single line with semicolons, but also ignore escaped semicolons
    user_input = user_input.replace("\\;","__WHITTLER_SEMICOLON_ESCAPE_W89UHA938F__")

    inputs = [cmd.strip().replace("__WHITTLER_SEMICOLON_ESCAPE_W89UHA938F__",";") for cmd in user_input.split(";")]

    ret = []
    for inp in inputs:
        # For now, just treat single and double quotes identically.
        inp = inp.replace("'","\"")

        if not inp.strip():
            continue

        if inp.find("\"") == -1:
            verb,*args = list(filter(None,inp.split(" ")))
        else:
            if inp.count("\"") % 2:
                wprint("Failed to parse quoted input.")
                return None
            quote_regions = inp.split("\"")
            # Obviously we're not going to support nested quotes, so every other quote_region will be
            # a string contained by quotes. Ideally I should be using the shlex library for this but
            # I'm too lazy.
            all_args = []
            for i in range(len(quote_regions)):
                if not i%2:
                    all_args.extend(list(filter(None,quote_regions[i].split(" "))))
                else:
                    all_args.append(quote_regions[i])
            verb,*args = all_args
        try:
            redirect_to_file_index = args.index(">")
            if redirect_to_file_index != len(args)-2:
                wprint("Specify only a single file to which to redirect command output! (Maybe try absolute paths; quotes are supported.)")
                return None
            redirect = args[-1]
            args = args[:-2]
        except ValueError:
            redirect = None
        
        for i in range(len(args)):
            args[i] = args[i].replace("###_TRUFFLEHOG_SINGLE_QUOTE_###","\\'")
            args[i] = args[i].replace("###_TRUFFLEHOG_DOUBLE_QUOTE_###","\\\"")
        ret.append((verb,args,redirect))
    return ret

irrelevance_filters = []
already_marked_relevant = []
def play_elimination_game(resultdb, obj):
    global irrelevance_filters
    global already_marked_relevant
    game_actions = {
        1 : "mark as relevant",
        2 : "mark as irrelevant",
        3 : "mark as ambiguous",
        4 : "quit game",
        5 : "clear relevancy filters"
    }
    def filterfunc(result):
        for attr, value in irrelevance_filters:
            if result[attr].strip() == value.strip():
                return False
        return True
    while True:
        results = list(filter(filterfunc, obj.all_result_objects()))
        if not len(results):
            wprint("All results accounted for - Game Over :)")
            break
        random_result = random.choice(results)
        wprint("\n".join(["| "+line for line in random_result.show_view()[0].splitlines()]))
        question = ""
        for action_id, action_description in game_actions.items():
            question += f" {action_id} : {action_description}\n"
        question += "\naction? "
        answer = winput(question)
        try:
            answer = int(answer)
            if not answer in game_actions.keys():
                wprint("Unrecognized selection.")
                continue
        except ValueError:
            wprint("Unrecognized selection.")
            continue
        if answer == 4:
            wprint("\nEnding game, thanks for playing.")
            break
        elif answer == 3:
            wprint("\nOK, taking no action.")
            continue
        elif answer == 1:
            wprint("\nOK, taking no action.")
            continue # todo
            relevant_attr = select_attribute(resultdb, "Which specific value makes this result definitely relevant? ")
            irrelevance_filters.append((relevant_attr,random_result[relevant_attr]))
            wprint("OK, I'll ignore results with that value for that attribute for the rest of the game.")
            continue
        elif answer == 5:
            wprint("Currently, you have relevancy filters on the following attributes:")
            for attr, value in irrelevance_filters:
                wprint(attr)
            confirm = winput("Are you sure you want to clear these relevancy filters? (Y/n) ")
            if confirm.lower() == "n":
                irrelevance_filters = []
                wprint("Cleared.")
                continue
        # answer must be 2 (irrelevant) past this point.
        wprint("\nOK, this result will be marked as irrelevant.\n")
        random_result.mark_irrelevant()
        num_results_eliminated = 1
        wprint("I can try to find other results similar to this one through various heuristics, and mark them irrelevant as well.\n")

        search_actions = {
            1 : "contains this specific value",
            2 : "contains a substring of this value (case-insensitive)",
            3 : "loosely resembles this value (fuzzy string matching)",
            4 : "never mind"
        }
        initial_str = "OK. To start, which"
        again_str = "Search for similar results? (y/N) "
        while True:
            go_again = winput(again_str)
            if go_again is None:
                wprint("Unrecognized input.")
                continue
            again_str = "Search again? (y/N)"
            if not go_again.strip() or go_again.lower() != "y":
                break
            wprint()
            wprint("\n".join(["| "+line for line in random_result.show_view()[0].splitlines()]))
            wprint()
            problematic_attr = select_attribute(resultdb, f"{initial_str} attribute's value makes this result irrelevant? ")
            initial_str = "Which"
            if not problematic_attr:
                wprint("Unrecognized selection.")
                continue
            wprint(f"\n\nFor reference, here is the value of the \"{problematic_attr}\" attribute for this result:")
            wprint("_"*Config.MAX_OUTPUT_WIDTH)
            wprint(random_result[problematic_attr])
            wprint("^"*Config.MAX_OUTPUT_WIDTH)

            question = f"\nI can search for results where the \"{problematic_attr}\" attribute:\n"
            for action_id, action_description in search_actions.items():
                question += f" {action_id} : {action_description.format(problematic_attr)}\n"
            question += "\naction? "
            answer = winput(question)
            try:
                answer = int(answer)
                if not answer in search_actions.keys():
                    wprint("Unrecognized selection.")
                    continue
            except ValueError:
                wprint("Unrecognized selection.")
                continue
            generate_report = True
            if answer == 1:
                irrelevant_resultlist = resultdb.categorized_results[problematic_attr][random_result[problematic_attr]]
                num_results_eliminated = len(irrelevant_resultlist)
                irrelevant_resultlist.mark_irrelevant()
            elif answer == 2:
                substring = winput("What is the substring you would like to filter by? ")
                if substring.lower() not in random_result[problematic_attr].lower():
                    wprint(f"Sorry, that is not a case-insensitive substring of this result's \"{problematic_attr}\" attribute...")
                    continue
                irrelevant_resultlist = resultdb.categorized_results[problematic_attr][random_result[problematic_attr]]
                for result in resultdb.results:
                    if substring.lower() in result[problematic_attr].lower():
                        result.mark_irrelevant()
                        num_results_eliminated += 1
            elif answer == 3:
                group_ptr = group_interactive(resultdb, problematic_attr, random_result[problematic_attr])
                if group_ptr is None:
                    continue
                irrelevant_resultdict = group_ptr.give_pointed_object()
                num_results_eliminated = len(irrelevant_resultdict.all_result_objects())
                irrelevant_resultdict.mark_irrelevant()
            elif answer == 4:
                generate_report = False
            
            if generate_report:
                excitement = num_results_eliminated//100
                report = f"\n > Eliminating {num_results_eliminated} results{'.' if not excitement else ''}{'!'*excitement}\n"
                if excitement >= 3:
                    report = report.upper()
                wprint(report)



import setuptools

with open("README.md", "r") as f:
    long_description = f.read()

setuptools.setup(
    name='Whittler',
    scripts=['Whittler/Whittler.py'],
    packages=setuptools.find_packages(),
    version='1.0.7',
    license='MIT',
    description="A machine-learning-capable modular shell for reducing large datasets (especially code static analysis tool outputs)",
    author="Ben Ege Izmirli",
    author_email="benegegalaxy@gmail.com",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/BenEgeIzmirli/Whittler",
    download_url="https://github.com/BenEgeIzmirli/Whittler/archive/1.0.7.tar.gz",
    # keywords = [], # TODO
    install_requires = [
        'numpy'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Topic :: Security",
        "Topic :: Database",
        "Topic :: Utilities",
        "Topic :: Text Processing :: Linguistic"
    ],
)


import zlib

# based on https://stackoverflow.com/questions/479218/how-to-compress-small-strings

class CompressedBytes(bytes):
    pass

class SingletonMeta(type):
    _instances = {}
    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(SingletonMeta, cls).__call__(*args, **kwargs)
        return cls._instances[cls]

class MemoryCompressor(metaclass=SingletonMeta):
    def __init__(self):
        # wbits=-15 gives a raw output stream with no header or checksum, with a 2**15 byte window.
        self.compressor = zlib.compressobj(wbits=-15)
        self.decompressor = zlib.decompressobj(wbits=-15)
        self.junk_offset = 0

        self.trainee_compression_callbacks = set()
        self.training_mode = True
        self.total_train_count = 1000

    def train(self, training_str, training_resultobj):
        self.trainee_compression_callbacks.add(training_resultobj)

        training_bytes = training_str.encode('utf-8')
        self.junk_offset += len(training_bytes)

        # run the training line through the compressor and decompressor
        self.junk_offset -= len(self.decompressor.decompress(self.compressor.compress(training_bytes)))

        # use Z_SYNC_FLUSH. A full flush seems to detrain the compressor, and 
        # not flushing wastes space.
        self.junk_offset -= len(self.decompressor.decompress(self.compressor.flush(zlib.Z_SYNC_FLUSH)))

        if len(self.trainee_compression_callbacks) >= self.total_train_count and training_resultobj not in self.trainee_compression_callbacks:
            self.disable_training_mode()

        return training_str
    
    def disable_training_mode(self):
        self.training_mode = False
        for resultobj in self.trainee_compression_callbacks:
            frozenstate = resultobj._frozen
            resultobj._frozen = False
            # This will get the raw string value from each attribute, then invoke the memory compression through __setitem__
            for attr in resultobj.ATTRIBUTES:
                if attr in resultobj:
                    rawval = dict.__getitem__(resultobj, attr)
                    if type(rawval) != CompressedBytes:
                        resultobj[attr] = rawval
            resultobj._frozen = frozenstate

    # Takes str, returns CompressedBytes
    def compress(self,s):
        compressor = self.compressor.copy()
        return CompressedBytes(compressor.compress(s.encode('utf-8'))+compressor.flush())

    # Takes CompressedBytes, returns str
    def decompress(self,b):
        decompressor = self.decompressor.copy()
        return (decompressor.decompress(b)+decompressor.flush())[self.junk_offset:].decode('utf-8')


from Whittler.classes.Result import RelevanceInterface, Result, ResultDictContainer, RelevanceFilteredResultList, ValueLengthSortedResultDict
from Whittler.classes.NestedObjectPointer import NestedObjectPointer, NestedObjectPointerInterface
from Whittler.classes.input_utils import wprint
from Whittler.config import Config
from collections import OrderedDict
import time
import os
import json
import numpy as np
import pickle
import gzip

try:
    from pyxdameraulevenshtein import normalized_damerau_levenshtein_distance
except ImportError:
    print("Warning: pyxDamerauLevenshtein module not detected, fuzzy grouping logic will be impaired.")
    print("         To install it, run 'pip install pyxDamerauLevenshtein'.")
    normalized_damerau_levenshtein_distance = lambda reference, values: np.array([1 for v in values])
import difflib


class ResultDatabase(RelevanceInterface):
    
    def __init__(self, result_class=Result):
        assert issubclass(result_class, Result)
        NestedObjectPointerInterface.__init__(self)
        RelevanceInterface.__init__(self)
        self.result_class = result_class
        self.results = RelevanceFilteredResultList()
        self.categorized_results = ResultDictContainer(self)
        self.grouped_results = ResultDictContainer(self)
        
        self.root_pointer = NestedObjectPointer(self)
        self.current_pointer = self.root_pointer.copy()
        self.context_pointers = OrderedDict()
        self._construct_view_cache = None
        self._construct_view_cache_pointer = None
        self._construct_view_cache_params = None
    
    def __getitem__(self, nestedobjectpointer):
        assert nestedobjectpointer.base_object is self
        return nestedobjectpointer.give_pointed_object()

    def add_result(self, result, lookup_set=None):
        assert isinstance(result,self.result_class)
        if lookup_set is None:
            if result in self.results:
                return
        else:
            result_hash = hash(result)
            if result_hash in lookup_set:
                return
            lookup_set.add(result_hash)
        result._frozen = True
        self.results.append(result)
        for attr in self.result_class.ATTRIBUTES:
            self.categorized_results[attr][result[attr]].append(result)


    #########################
    #  Navigation functions
    #

    def construct_view(self, pointer=None, limit=None, show_irrelevant=False, sort_by=None, sort_numeric=False, sort_reverse=False):
        if pointer is None:
            pointer = self.current_pointer
        if self._construct_view_cache_pointer is not None and self._construct_view_cache_pointer == pointer:
            if self._construct_view_cache_params is not None and (limit,show_irrelevant,sort_by,sort_numeric,sort_reverse)==self._construct_view_cache_params:
                return self._construct_view_cache
        obj = pointer.give_pointed_object()
        viewstr, ptrdict = obj.show_view(pointer_to_me=pointer, limit=limit, show_irrelevant=show_irrelevant, sort_by=sort_by, sort_numeric=sort_numeric, sort_reverse=sort_reverse)
        self._construct_view_cache_pointer = pointer.copy()
        self._construct_view_cache = (viewstr, ptrdict)
        self._construct_view_cache_params = (limit,show_irrelevant,sort_by,sort_numeric,sort_reverse)
        return self._construct_view_cache
    
    def update_current_view(self, limit=None, sort_by=None, sort_numeric=False, sort_reverse=False):
        viewstr, ptrcontext = self.construct_view(self.current_pointer, limit=limit, sort_by=sort_by, sort_numeric=sort_numeric, sort_reverse=sort_reverse)
        self.context_pointers = ptrcontext
        return viewstr
    
    def navigate_view(self, pointer=None, limit=None, sort_by=None, sort_numeric=False, sort_reverse=False):
        if pointer is None:
            self.current_pointer = self.root_pointer.copy()
        else:
            self.current_pointer = pointer.copy()
        return self.update_current_view(limit=limit, sort_by=sort_by, sort_numeric=sort_numeric, sort_reverse=sort_reverse)
    

    #######################
    #  Grouping functions
    #

    def group_attribute(self, attrname):
        raise NotImplementedError()
    
    @staticmethod
    def compute_distances(reference,values,exp=None):
        if exp is None:
            exp = Config.SIMILARITY_EXPONENT

        dh_distances = 1-normalized_damerau_levenshtein_distance(reference, values)
        
        sm = difflib.SequenceMatcher()
        sm.set_seq2(reference)
        sm_distances = []
        for val in values:
            sm.set_seq1(val)
            sm_distances.append(sm.ratio())
        sm_distances = np.array(sm_distances)
        
        dh_exp = np.power(dh_distances, exp)
        sm_exp = np.power(sm_distances, exp)
        dist_sum = dh_exp+sm_exp
        return np.power(dist_sum,1/exp)
    
    def find_similar_results(self, attrname, groupval, similarity_threshold=None):
        if similarity_threshold is None:
            similarity_threshold = self.Config.SIMILARITY_THRESHOLD
        if not groupval.strip():
            all_results = [res for res in list.__iter__(self.results) if hasattr(res,attrname) and res[attrname].strip()]
        else:
            all_results = [res for res in list.__iter__(self.results)]
        attrarray = np.array([res[attrname].strip() for res in all_results])
        normalized_distances = self.compute_distances(groupval, attrarray)
        return [res for res,_ in filter(lambda tup:tup[1]>similarity_threshold, zip(all_results, normalized_distances))]
    
    def group_attribute_by_value(self, attrname, groupval):
        similar_results = self.find_similar_results(attrname, groupval)
        self.register_grouped_results(attrname, groupval, similar_results)

    def register_grouped_results(self, attrname, groupval, result_group):
        for result in result_group:
            self.grouped_results[groupval][result[attrname]].append(result)
        self.grouped_results.flush_pointer_cache()


    #################################
    #  Input file parsing functions
    #

    def parse_from_file(self,fname,hash_cache=set()):
        result_dict_list = self.result_class.give_result_dict_list(fname)
        last_report = time.time()
        start = last_report
        ct = 0
        parsing_str = f"PARSING {fname} : "
        longest_str_length = 0
        for resultdict in result_dict_list:
            cur_time = time.time()
            if cur_time-last_report > 5:
                print_str = f"{parsing_str}{(int((ct/len(result_dict_list)*100)))}% done ({ct} out of {len(result_dict_list)})\r"
                if len(print_str)>longest_str_length:
                    longest_str_length = len(print_str)
                wprint(print_str)
                last_report = cur_time
            resultdict["whittler_filename"] = fname
            self.add_result(self.result_class(resultdict), lookup_set=hash_cache)
            ct += 1
        if longest_str_length:
            timing = "{:.2f}".format(time.time()-start)
            final_report = f"{parsing_str}Done (took {timing}s)"
            wprint(final_report + " "*(len(final_report)-max(Config.MAX_OUTPUT_WIDTH,longest_str_length)))

    def parse_from_directory(self,dirname,hash_cache=set()):
        files = os.listdir(dirname)
        last_report = time.time()
        start = last_report
        ct = 0
        parsing_str = f"FILES PARSED: "
        longest_str_length = 0
        wprint(f"Parsing from {dirname} ...")
        for output_file in files:
            cur_time = time.time()
            if cur_time-last_report > 5:
                print_str = f"{parsing_str}{(int((ct/len(files)*100)))}% ({ct} out of {len(files)})\r"
                if len(print_str)>longest_str_length:
                    longest_str_length = len(print_str)
                wprint(print_str)
                last_report = cur_time
            try:
                self.parse_from_file(dirname+"/"+output_file, hash_cache=hash_cache)
            except PermissionError as e:
                wprint(f"WARNING: failed to open {dirname+'/'+output_file}: {e}")
            ct += 1
        timing = "{:.2f}".format(time.time()-start)
        final_report = f"Parsing from {dirname} done (took {timing}s)"
        wprint(final_report + " "*(len(final_report)-max(Config.MAX_OUTPUT_WIDTH,longest_str_length)))
    
    def parse_from_export(self,fname,hash_cache=set()):
        importing_str = f"IMPORTING {os.path.basename(fname)} ... "
        last_report = time.time()
        start = last_report
        wprint(f"{importing_str}", end='\r')
        pickle_import = False
        try:
            try:
                with gzip.GzipFile(fname, "rb") as f:
                    results = pickle.load(f)
                pickle_import = True
            except pickle.UnpicklingError:
                try:
                    with gzip.GzipFile(fname, "rb") as f:
                        results = json.loads(f.read().decode('utf-8'))
                except json.decoder.JSONDecodeError:
                    raise Exception("Failed to import file as either binary (pickle) or JSON data.")
        except:
            raise
        # If we're importing from JSON, we need to make sure that the data keys are compatible with the specified module
        if not pickle_import:
            first_result_keys = set(results[0].keys())
            for result in results:
                if first_result_keys ^ set(result.keys()): # check symmetric difference, basically ensure that they're equal
                    raise Exception(f"invalid Whittler export file {fname} ... all results dicts in the JSON must have the same "+\
                                    f"set of keys, but the keys in the result:\n{result}\ndid not match the expected keys:\n"+\
                                    f"{first_result_keys}")
            if not isinstance(self.result_class.ATTRIBUTES, list):
                self.result_class.ATTRIBUTES = []
            attrs = self.result_class.ATTRIBUTES
            for key in results[0].keys(): # don't use first_result_keys because the json is ordered whereas sets are not
                if key not in attrs:
                    attrs.append(key)
        ct = 0
        biggest_status_str_len = 0
        for result in results:
            cur_time = time.time()
            if cur_time-last_report > 5:
                status_str = f"{importing_str}{(int((ct/len(results)*100)))}% done ({ct} out of {len(results)})"
                if len(status_str) > biggest_status_str_len:
                    biggest_status_str_len = len(status_str)
                wprint(status_str, end='\r')
                last_report = cur_time
            if pickle_import:
                self.add_result(result, lookup_set=hash_cache)
            else:
                self.add_result(self.result_class(result), lookup_set=hash_cache)
            ct += 1
        tot_time = "{:.2f}".format(time.time()-start)
        done_str = f"{importing_str}Done (took {tot_time}s)"
        wprint(done_str+" "*max(biggest_status_str_len,Config.MAX_OUTPUT_WIDTH-len(done_str)))


    #######################################
    #  RelevanceInterface implementations
    #
    
    def real_iter_values(self):
        return self.results.real_iter_values()
    

    #################################################
    #  NestedObjectPointerInterface implementations
    #
    
    def enumerate_child_pointers(self, pointer_to_me):
        ret = OrderedDict()
        
        ptr = self.root_pointer.copy()
        ptr.access_property("categorized_results")
        ret["categorized_results"] = ptr
        
        ptr = self.root_pointer.copy()
        ptr.access_property("grouped_results")
        ret["grouped_results"] = ptr
        
        self._cached_pointers = ret
        return ret
    
    def size(self):
        return self.results.size()
    
    def all_result_objects(self):
        return self.results.all_result_objects()
    
    def show_view(self, pointer_to_me=None, ct=0, limit=None, show_irrelevant=False, sort_by=None, sort_numeric=False, sort_reverse=False):
        s = "\n"
        s += "+=================================+\n"
        s += "|  {:6d} total relevant results  |\n".format(len(self.results))
        s += "+=================================+\n"
        ret = OrderedDict()
        for childname, childptr in self.give_child_pointers(self.root_pointer).items():
            s += f"\n{childname}:\n"
            childobj = childptr.give_pointed_object()
            s2, resultdict = childobj.show_view(childptr, ct=ct, limit=limit, show_irrelevant=show_irrelevant, sort_by=sort_by, sort_numeric=sort_numeric, sort_reverse=sort_reverse)
            s += s2
            for k,v in resultdict.items():
                ret[k] = v
            ct += len(resultdict)
        return (s, ret)

    def export(self):
        return self.results.export()
    


import os
import sys

WHITTLER_DIRNAME = os.path.dirname(os.path.realpath(__file__))
# print(os.path.dirname(WHITTLER_DIRNAME))
sys.path.append(os.path.dirname(WHITTLER_DIRNAME))
try:
    sys.path.remove(WHITTLER_DIRNAME)
except ValueError: # Already removed
    pass

from Whittler.classes.ResultDatabase import ResultDatabase
from Whittler.classes.Result import Result, RelevanceFilteredResultList
from Whittler.classes.input_utils import *
from Whittler.classes.MemoryCompressor import MemoryCompressor
import importlib
import datetime
import argparse
from pathlib import Path
import json
import pickle
import gzip


def main_loop(resultdb):
    global redirect_file, global_redirect_file, cached_commands
    sort_by = None
    sort_numeric = False
    sort_reverse = False
    while True:
        if not redirect_file is None:
            redirect_file.close()
            redirect_file = None
        resultdb.update_current_view(sort_by=sort_by, sort_numeric=sort_numeric, sort_reverse=sort_reverse)
        if len(cached_commands):
            user_input = cached_commands[0]
            cached_commands = cached_commands[1:]
            wprint(f"Whittler > {' '.join([user_input[0]]+user_input[1])}")
        else:
            user_input = parse_user_input(winput("Whittler > "))
            if user_input is None:
                continue
            if len(user_input) > 1:
                cached_commands = user_input[1:]
            user_input = user_input[0]
        if user_input is None:
            continue
        verb,args,redirect = user_input
        if verb.startswith("#"):
            continue

        if not redirect is None:
            try:
                redirect_file = open(redirect, "w+", encoding="utf-8")
            except PermissionError:
                wprint("Failed to open the specified file, maybe try an absolute path? (FYI, quotes are supported.)")

        if verb == "help":
            print_help()
            continue


        ########################
        #  Navigation commands
        #

        elif verb == "show":
            limit = get_int_from_args(args)
            limit = limit if limit else None # get_int_from_args returns False if no value was supplied for the arg
            wprint(resultdb.construct_view(limit=limit, sort_by=sort_by, sort_numeric=sort_numeric, sort_reverse=sort_reverse)[0])
            continue
        elif verb == "showall":
            limit = get_int_from_args(args)
            limit = limit if limit else None # get_int_from_args returns False if no value was supplied for the arg
            wprint(resultdb.construct_view(limit=limit, show_irrelevant=True, sort_by=sort_by, sort_numeric=sort_numeric, sort_reverse=sort_reverse)[0])
            continue
        elif verb == "dig":
            ptr = get_ptr_from_id_arg(resultdb, args)
            if ptr is False:
                wprint("no [attr] specified, no digging performed.")
                continue
            if ptr is None:
                continue
            sort_by = None
            sort_numeric = False
            sort_reverse = False
            resultdb.navigate_view(ptr)
            continue
        elif verb == "up":
            if resultdb.current_pointer.is_base_pointer():
                wprint("Already at root context.\n")
                continue
            sort_by = None
            sort_numeric = False
            sort_reverse = False
            resultdb.current_pointer.go_up_level()
            # We want to pop out from the categorized_results or grouped_results contexts implicitly.
            if len(resultdb.current_pointer.path) == 1:
                resultdb.current_pointer.go_up_level()
            continue
        elif verb == "top":
            while not resultdb.current_pointer.is_base_pointer():
                resultdb.current_pointer.go_up_level()
            sort_by = None
            sort_numeric = False
            sort_reverse = False
            continue
        elif verb == "dump":
            limit = get_int_from_args(args)
            limit = limit if limit else None # get_int_from_args returns False if no value was supplied for the arg
            wprint(resultdb.results.show_view(limit=limit)[0])
        elif verb == "dumpall":
            limit = get_int_from_args(args)
            limit = limit if limit else None # get_int_from_args returns False if no value was supplied for the arg
            # todo
            wprint(resultdb.results.show_view(limit=limit)[0])
        elif verb == "exit":
            sys.exit(0)


        ####################################
        #  Data model interaction commands
        #
        
        elif verb == "irrelevant":
            ptr = get_ptr_from_id_arg(resultdb, args)
            if ptr is False:
                ptr = resultdb.current_pointer
            elif ptr is None:
                continue
            obj = ptr.give_pointed_object()
            obj.mark_irrelevant()
            continue
        elif verb == "relevant":
            ptr = get_ptr_from_id_arg(resultdb, args)
            if ptr is False:
                ptr = resultdb.current_pointer
            elif ptr is None:
                continue
            obj = ptr.give_pointed_object()
            obj.mark_relevant()
            continue
        elif verb == "group":
            ptr = get_ptr_from_id_arg(resultdb, args)
            if ptr is False:
                wprint("need to specify an [id], the value of which to group by.")
                continue
            if ptr is None:
                continue
            obj = ptr.give_pointed_object()
            if isinstance(obj, resultdb.result_class):
                groupattr = get_attrname_from_attribute_arg(resultdb, args, attr_arg_position=1)
                if groupattr is False:
                    groupattr = select_attribute(resultdb, "Which attribute of this result would you like to group by? ")
                elif groupattr is None:
                    continue
                groupval = obj[groupattr]
            elif isinstance(obj, RelevanceFilteredResultList):
                groupval = ptr.go_up_level().value
                groupattr = ptr.go_up_level().value
            else:
                raise NotImplementedError()
            group_interactive(resultdb, groupattr, groupval)
            continue
        elif verb == "game":
            ptr = get_ptr_from_id_arg(resultdb, args)
            if ptr is False:
                obj = resultdb.results
            elif ptr is None:
                continue
            else:
                obj = ptr.give_pointed_object()
            play_elimination_game(resultdb, obj)
            continue
        elif verb == "filter":
            if not len(args):
                wprint("Need to provide a string to filter by.")
                continue
            filter_str = args[0]
            filtered_attr = get_attrname_from_attribute_arg(resultdb, args, attr_arg_position=1)
            if filtered_attr is False:
                filtered_attr = select_attribute(resultdb, "Which attribute would you like to filter by? ")
            elif filtered_attr is None:
                continue
            do_inverse = winput("Mark all matches as irrelevant (Y)? Or mark all non-matches as irrelevant (n)? ")
            if not do_inverse.strip() or do_inverse.lower() == "y":
                do_inverse = False
            else:
                do_inverse = True
            ct = 0
            for result in resultdb.results:
                if not do_inverse:
                    result_matches = filter_str.lower() in result[filtered_attr].lower()
                else:
                    result_matches = filter_str.lower() not in result[filtered_attr].lower()
                if result_matches:
                    result.mark_irrelevant()
                    ct += 1
            wprint(f"Marked {ct} results as irrelevant using the filter.")
            continue

        
        
        #####################
        #  Output commands
        #
        
        elif verb == "quiet":
            quieted_attr = get_attrname_from_attribute_arg(resultdb, args)
            if quieted_attr is False:
                quieted_attr = select_attribute(resultdb, "Which attribute would you like to suppress in output? ")
            elif quieted_attr is None:
                continue
            resultdb.result_class.SILENCED_ATTRIBUTES.add(quieted_attr)
            continue
        elif verb == "unquiet":
            attrs = list(resultdb.result_class.SILENCED_ATTRIBUTES)
            if not len(attrs):
                wprint("No silenced attributes to unquiet.")
                continue
            quieted_attr = get_attrname_from_attribute_arg(resultdb, args)
            if quieted_attr is False:
                for i in range(len(attrs)):
                    wprint(f" {i} : {attrs[i]}")
                wprint()
                index = int(winput("Which attribute would you like to un-suppress in output? "))
                quieted_attr = attrs[index]
            elif quieted_attr is None:
                continue
            elif quieted_attr not in attrs:
                wprint("That attribute was not silenced, no action taken.")
                continue
            resultdb.result_class.SILENCED_ATTRIBUTES.remove(quieted_attr)
            continue
        elif verb == "solo":
            solo_attr = get_attrname_from_attribute_arg(resultdb, args)
            if solo_attr is False:
                solo_attr = select_attribute(resultdb, "Which attribute would you like to print exclusively in output? ")
            elif solo_attr is None:
                continue
            resultdb.result_class.SOLO_ATTRIBUTE = solo_attr
            continue
        elif verb == "SOLO":
            solo_attr = get_attrname_from_attribute_arg(resultdb, args)
            if solo_attr is False:
                solo_attr = select_attribute(resultdb, "Which attribute would you like to print exclusively in output? ")
            elif solo_attr is None:
                continue
            resultdb.result_class.SUPER_SOLO_ATTRIBUTE = solo_attr
            continue
        elif verb == "unsolo":
            resultdb.result_class.SOLO_ATTRIBUTE = None
            resultdb.result_class.SUPER_SOLO_ATTRIBUTE = None
            continue
        elif verb == "sort":
            if not len(args):
                wprint("Need a column name or attribute value to sort by.")
            sort_by = args[0]
            sort_numeric = False
            sort_reverse = False
            continue
        elif verb == "sortn":
            if not len(args):
                wprint("Need a column name or attribute value to sort by.")
            sort_by = args[0]
            sort_numeric = True
            sort_reverse = False
            continue
        elif verb == "rsort":
            if not len(args):
                wprint("Need a column name or attribute value to sort by.")
            sort_by = args[0]
            sort_numeric = False
            sort_reverse = True
            continue
        elif verb == "rsortn":
            if not len(args):
                wprint("Need a column name or attribute value to sort by.")
            sort_by = args[0]
            sort_numeric = True
            sort_reverse = True
            continue
        elif verb == "history":
            wprint()
            for cmd in command_history:
                wprint(cmd)
            wprint()
            continue
        elif verb == "width":
            if not len(args):
                wprint("Need a column name or attribute value to sort by.")
            try:
                new_width = int(args[0])
            except:
                wprint(f"Failed to parse {args[0]} as an integer.")
                continue
            Config.MAX_OUTPUT_WIDTH = new_width
            continue
        elif verb == "exportjson":
            if not len(args):
                wprint("Need a filename to export to.")
                continue
            try:
                fname = args[0]
                ptr = get_ptr_from_id_arg(resultdb, args, id_arg_position=1)
                if ptr is False:
                    obj = resultdb
                elif ptr is None:
                    continue
                else:
                    obj = ptr.give_pointed_object()
                resultlist = obj.export()
                try:
                    if os.path.isfile(fname):
                        answer = winput("WARNING: file already exists. Override? (N/y) ").strip().lower()
                        if not (answer == "y" or answer == "yes"):
                            wprint("Aborting export, no files written.")
                            continue
                        wprint("Overwriting file...")
                    with gzip.GzipFile(fname,"wb") as f:
                        for chunk in json.JSONEncoder().iterencode(resultlist):
                            f.write(chunk.encode('utf-8'))
                    wprint(f"Export success, compressed JSON output written to {fname}.")
                except PermissionError:
                    wprint("Failed to open the specified file, maybe try an absolute path? (FYI, quotes are supported.)")
                continue
            except Exception as e:
                wprint(f"Exception encountered while exporting: {e}")
        elif verb == "export":
            if not len(args):
                wprint("Need a filename to export to.")
                continue
            try:
                fname = args[0]
                ptr = get_ptr_from_id_arg(resultdb, args, id_arg_position=1)
                if ptr is False:
                    # no specific object (result, or row, etc) was specified - export everything
                    obj = resultdb
                elif ptr is None:
                    wprint("Couldn't find the specified result, no action taken.")
                    continue
                else:
                    obj = ptr.give_pointed_object()
                resultlist = obj.all_result_objects()
                try:
                    if os.path.isfile(fname):
                        answer = winput("WARNING: file already exists. Override? (N/y) ").strip().lower()
                        if not (answer == "y" or answer == "yes"):
                            wprint("Aborting export, no files written.")
                            continue
                        wprint("Overwriting file...")
                    with gzip.GzipFile(fname,"wb") as f:
                        pickle.dump(resultlist, f)
                    wprint(f"Export success, compressed serialized output written to {fname}.")
                except PermissionError:
                    wprint("Failed to open the specified file, maybe try an absolute path? (FYI, quotes are supported.)")
                continue
            except Exception as e:
                wprint(f"Exception encountered while exporting: {e}")

        else:
            wprint("Unrecognized command.\n")
            continue

        
HOME_DIRECTORY = str(Path.home())
WHITTLER_DIRECTORY = HOME_DIRECTORY+"/.whittler"
if not os.path.isdir(WHITTLER_DIRECTORY):
    os.mkdir(WHITTLER_DIRECTORY,mode=0o770)

result_classes = {}
for fname in filter(lambda s: not s.startswith("_") , os.listdir(WHITTLER_DIRNAME+"/modules")):
    module = importlib.import_module(f"Whittler.modules.{fname[:fname.index('.')]}")
    for clsname in dir(module):
        if clsname.startswith("__"):
            continue
        cls = getattr(module, clsname)
        if isinstance(cls, type) and issubclass(cls, Result) and not cls is Result:
            result_classes[cls.FRIENDLY_NAME] = cls
            break


# to be populated if the --script or --scriptfile flags are specified
cached_commands = []

def main():
    global cached_commands
    # This is a patch for running the script with "python -m Whittler" - without this modification the help messages
    # think the script is named "__main__.py".
    old_argv0 = sys.argv[0]
    sys.argv[0] = old_argv0.replace("__main__","Whittler")

    parser = argparse.ArgumentParser(description="An interactive script to whittle down large datasets")

    # Reset the script name back to the old value, just in case.
    sys.argv[0] = old_argv0

    # Required args
    bargs = parser.add_argument_group("basic arguments")
    bargs.add_argument('--config',
                    help='the module to use to parse the specified tool output files.',
                    type=str, nargs=1, choices=list(result_classes.keys()), default=None, required=True)

    # Data ingestion args
    diargs = parser.add_argument_group("data ingestion arguments")
    #diargs = diargs.add_mutually_exclusive_group(required=True)
    diargs.add_argument('--file',
                        help='the tool output file to be parsed',
                        type=str, nargs='+', default='')
    diargs.add_argument('--dir',
                        help='the directory containing tool output files to be parsed',
                        type=str, nargs='+', default='')
    diargs.add_argument('--import_whittler_output',
                        help='consume and continue working with one or more files that were outputted by Whittler\'s "export" command',
                        type=str, nargs='+', default=None, metavar="FILE_OR_DIR")
    diargs.add_argument('--memory_compression',
                        help='enable in-memory compression for working with very large datasets',
                        action='store_true')

    # Output control args
    ocargs = parser.add_argument_group("output control arguments")
    ocargs.add_argument('--log_output',
                        help='a file to which all output in this session will be logged (default: a new file in the '+\
                            '.whittler folder in your home directory)',
                        type=str, nargs="?", default=None, metavar="FILENAME",
                        const=WHITTLER_DIRECTORY+'/{date:%Y-%m-%d_%H-%M-%S}_log.txt'.format( date=datetime.datetime.now() ) )
    ocargs.add_argument('--log_command_history',
                        help='a file in which to record the command history of this session, in a format that can be '+\
                            'imported and re-run by the --scriptfile flag (default: a new file in the .whittler folder in '+\
                            'your home directory)',
                        type=str, nargs="?", default=None, metavar="FILENAME",
                        const=WHITTLER_DIRECTORY+'/{date:%Y-%m-%d_%H-%M-%S}_command_log.txt'.format( date=datetime.datetime.now() ))

    # Scripting arguments
    sargs = parser.add_argument_group("scripting arguments")
    sargs.add_argument('--script',
                    help='run a script specified with a string on the command line, with each command separated by semicolons '+\
                            '(backslash-escape for a literal semicolon)',
                    type=str, nargs=1, default=None, metavar="SCRIPT_STRING")
    sargs.add_argument('--scriptfile',
                    help='run a script provided in a file, with one command per line',
                    type=str, nargs=1, default=None, metavar="SCRIPT_FILE")

    args = parser.parse_args()

    if args.memory_compression:
        Config.MEMORY_COMPRESSION = True

    if not args.log_output is None:
        logdir = args.log_output[0] if isinstance(args.log_output,list) else args.log_output
        try:
            set_global_redirect_file(open(logdir,"w+", encoding="utf-8"))
        except PermissionError:
            wprint("Lacking permissions to write to the specified all-output log file.")
            sys.exit(1)
        get_global_redirect_file().write(" ".join(sys.argv)+"\n\n")
    if not args.log_command_history is None:
        logcmddir = args.log_command_history[0] if isinstance(args.log_command_history,list) else args.log_command_history
        try:
            set_command_redirect_file(open(logcmddir,"w+", encoding="utf-8"))
        except PermissionError:
            wprint("Lacking permissions to write to the specified command history log file.")
            sys.exit(1)
    if args.config is None:
        parser.print_help()
        sys.exit(1)
    try:
        resultdb = ResultDatabase(result_classes[args.config[0]])
        if not args.dir and not args.file and not args.import_whittler_output:
            parser.print_help()
            sys.exit(1)
        wprint("\nWelcome to the Whittler shell. Type \"help\" for a list of commands.\n")
        if args.script and args.scriptfile:
            wprint("Error: the --script and --scriptfile flags cannot both be specified at once.\n")
            parser.print_help()
            sys.exit(1)
        if args.script or args.scriptfile:
            if args.script:
                wprint(f"Using script from command line.")
                cached_commands = parse_user_input(args.script[0])
            else:
                wprint(f"Importing script from file {args.scriptfile[0]} .")
                with open(args.scriptfile[0],"r") as f:
                    cached_commands = parse_user_input(" ; ".join(f.readlines()))
        hash_cache = set()
        if args.dir:
            wprint("Parsing files from provided directories...")
            wprint()
            for d in args.dir:
                resultdb.parse_from_directory(d, hash_cache=hash_cache)
        if args.file:
            wprint("Parsing provided files...")
            wprint()
            for fname in args.file:
                resultdb.parse_from_file(fname, hash_cache=hash_cache)
        if args.import_whittler_output:
            wprint("Importing provided files...")
            wprint()
            for import_target in args.import_whittler_output:
                if os.path.isdir(import_target):
                    import_target = [import_target+"/"+fname for fname in os.listdir(import_target)]
                else:
                    import_target = [import_target]
                for fname in import_target:
                    resultdb.parse_from_export(fname, hash_cache=hash_cache)
        
        if Config.MEMORY_COMPRESSION and Config.MemoryCompressor.training_mode:
            Config.MemoryCompressor.disable_training_mode()
        wprint("Done.\n")
        main_loop(resultdb)
    except:
        raise
    finally:
        print()
        if not command_redirect_file is None:
            fname = command_redirect_file.name.replace('\\','/')
            wprint(f"Saving logged command history to {fname} ...")
            command_redirect_file.close()
        if not redirect_file is None:
            fname = redirect_file.name.replace('\\','/')
            wprint(f"Saving pipe output to {fname} ...")
            redirect_file.close()
        goodbye_said = False
        if not global_redirect_file is None:
            fname = global_redirect_file.name.replace('\\','/')
            wprint(f"Saving logged output to {fname} ...")
            wprint("\nGoodbye.\n")
            goodbye_said = True
            global_redirect_file.close()
        if not goodbye_said:
            print("\nGoodbye.\n")
        

if __name__ == "__main__":
    main()



from Whittler.classes.Result import Result
import csv

class PSSAResult(Result):

    FRIENDLY_NAME = "pssa_csv"
    
    ATTRIBUTES = []

    @staticmethod
    def give_result_dict_list(fname):
        ret = []
        with open(fname, newline='') as csvfile:
            filecache = {}
            for row in csv.DictReader(csvfile):
                if not PSSAResult.ATTRIBUTES:
                    for key in row.keys():
                        PSSAResult.ATTRIBUTES.append(key)
                    PSSAResult.ATTRIBUTES.append("FullLine")
                    PSSAResult.ATTRIBUTES.append("FullLineContext")
                if "ScriptPath" in row and "Line" in row:
                    try:
                        path = row["ScriptPath"]
                        if path not in filecache:
                            with open(row["ScriptPath"],"r") as f:
                                filecache[path] = f.read()
                        fcontent = filecache[path]
                        flines = fcontent.splitlines()
                        lineno = int(row["Line"])-1
                        row["FullLine"] = flines[lineno].strip()
                        row["FullLineContext"] = "\n".join(flines[lineno-5: lineno+5])
                    except Exception as e:
                        if isinstance(e, IndexError):
                            print(f"WARNING: line number {lineno} too high for file with {len(flines)} lines: {path}")
                        row["FullLine"] = ""
                        row["FullLineContext"] = ""
                else:
                    row["FullLine"] = ""
                    row["FullLineContext"] = ""
                        
                ret.append(row)
        return ret